/*Program started on June 19, 2007

SAC to ASC Converter
Created by Kunnal Khiatani as a Research Assistant for Dr.Hrvoje Tkalcic at the RSES, Australian National University

Last Edited: June 20, 2007
Days worked on: June 19, 20
Re-edited on: September 13, 2007 because : Bug found. Some sac files would produce empty ascii files.
Problem solved using 0/1 reverse flag.

Status: Complete! Functional for both evenly spaced and non-evenly spaced SAC Files.

Notes:
Step 2: SAC to ASC (Complete)
Step 1: ZDF to SAC (Complete)

Usage:
sac2asc <sacFile> 0/1[Optional]

(1) sac2asc is the executable either sac2ascSPARC or sac2ascX86
(2) the sac file to be converted to ascii file
(3) Enter 1 if the sac file has been created on an X86 machine, or 0 if created on a SPARC Machine - default is 1 (X86)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LITTLE_ENDIAN 	0
#define BIG_ENDIAN 	1

/*SAC Definition Block - Defined in the Byte-order of the SAC Format*/
float	DELTA, B;
int	SACNPTS; 
/*End of SAC Definition Block*/


/*Function Definitions*/
int endian_check();
void swap_bytes(char *buffer,int samples,int size);
/*End of Function Definitions*/


/*Variable Definitions*/
int exitFlag = 0;
int byteSwapFlag = 0;
int endian = 1; //Default to Big Endian
char ascFileName[100];
char junk[1000];
float datapoint;
//float  datapointArray[500000];
int i = 0;
/*End of Variable Defintions*/


/*Start of Main Program*/
int main(int argc, char* argv[])
{
  char buffer[100];
  int *ptint;
  float *ptfloat;
  int reverse;


  /*Check for correct input parameters*/
  if (argc<2)
    {
      printf("Insufficient Parameters. Exiting...\n");
      exitFlag = 1;
    }
  else if (argc>3)
    {
      printf("Too Many Parameters. Exiting...\n");
      exitFlag = 1;
    }

 else if (argc==2)
 {
    reverse=1;		//Reverse handles for sac files that have been created on
 }			//X86 machines and hence dont need byteswapping.
			//Default assumption is that sac has been created on X86 machines
 else if (argc==3)
 {
  int argv_Value = atoi(argv[2]);

  if (argv_Value==1)
   reverse = 1;
  else if (argv_Value==0)
   reverse = 0;
 }

//printf("Reverse: %d\n",reverse);

  while(exitFlag!=1)
    {/* Main Body*/
      
      /*Sac File Reading*/
      FILE *sacInput, *ascOutput;
      sacInput = fopen(argv[1],"rb"); //Contains the SAC File

      if(sacInput == NULL)
	{ 
	  printf("Empty SAC File Found. \n");
	}
      else
	{
	  /*All File reading and writing, etc is done in this else loop*/
	  /*Little-Endian or Big-Endian check*/
	  endian = endian_check();
	  if (endian == 0)
	    {
	      //printf("Little Endian\n");
	      byteSwapFlag = 1;
		if(reverse==1)
		{byteSwapFlag = 0;}
	    }
	  else if (endian == 1)
	    {
	      //printf("Big Endian\n");
	      byteSwapFlag = 0;
	    }
	  
	  /*Get DELTA*/
	  fread(buffer, sizeof(char),4, sacInput);
	  if(byteSwapFlag == 1)
	    {
	      swap_bytes(buffer,1,4);
	    }
	  ptfloat = (float*)buffer;
	  DELTA = ptfloat[0];
	  //printf("Delta = %f\n",DELTA);
	  
 	  /*Get Junk*/
	  fread(junk,sizeof(char),16,sacInput);
	  //printf("%s\n",junk);

 	  /*Get Beginning Time (B)*/
	  fread(buffer, sizeof(char),4, sacInput);
	  if(byteSwapFlag == 1)
	    {
	      swap_bytes(buffer,1,4);
	    }
	  ptfloat = (float*)buffer;
	  B = ptfloat[0];
	  //printf("B = %f\n",B);
	 
	  /*Get Junk*/
	  fread(junk,sizeof(char),292,sacInput);
	  //printf("%s\n",junk);
	  
	  /*Get NPTS*/
	  fread(buffer, sizeof(char),4,sacInput);
	  if(byteSwapFlag == 1)
	    {
	      swap_bytes(buffer,1,4);
	    }
	  ptint = (int*)buffer;
	  SACNPTS = ptint[0];
	  //printf("NPTS = %d\n", SACNPTS);
	  
	  fread(junk, sizeof(char),312, sacInput);
	  
	  //fread(&datapoint, sizeof(float),1,sacInput);
	  //printf("%f\n",datapoint);
	  
	  /*ASCII File Creation and Writing*/
	  char tempAscName[100];
	  strcpy(ascFileName,"");
	  strcpy(ascFileName,argv[1]);
	  //sscanf(ascFileName, "%s", tempAscName);
	  //int charsToCopy = (strlen(tempAscName))-3;
	  // printf("%d\n",charsToCopy);
	  //strcpy(ascFileName,"");
	  //printf("%s\n",ascFileName);
	  //strncat(ascFileName, tempAscName, charsToCopy);
	  //printf("%s\n",ascFileName);  
	  strcat(ascFileName, ".asc");
	  ascOutput = fopen(ascFileName, "w");
	  
	  float dataPointArray[SACNPTS];
	  double  timeVariableArray[SACNPTS];//double

	  if(DELTA!= -12345.0) /*Copies Info from Evenly Spaced SAC File*/
	    {
	      double sum = (double) B;//double
	      for(i=0;i<SACNPTS;i++)
		{
		  //fread(&datapoint, sizeof(float),1,sacInput);
		  fread(buffer,sizeof(char),4,sacInput);
		  if(byteSwapFlag == 1)
		    {
		      swap_bytes(buffer,1,4);
		    }
		  ptfloat = (float*)buffer;
		  dataPointArray[i] = ptfloat[0];
		  fprintf(ascOutput,"%e\t%f\n",sum,dataPointArray[i]);
		  sum = sum + (double) DELTA; //(double)
		  //B = B + DELTA;
		}
	    }
	  else /*Copies Info from Unevenly Spaced SAC File */
	    {
	      
	      for (i=0;i<SACNPTS;i++)
		{
		  fread(buffer,sizeof(char),4,sacInput);
		  if(byteSwapFlag == 1)
		    {
		      swap_bytes(buffer,1,4);
		    }
		  ptfloat = (float*)buffer;
		  dataPointArray[i] = ptfloat[0];
		  //fprintf(ascOutput,"%e\t%f\n",TimeVariable,datapoint);
		}//end for - First Segment of SAC received - Amplitudes
	      for (i=0; i<SACNPTS;i++)
		{
		  fread(buffer,sizeof(char),4,sacInput);
		  if(byteSwapFlag == 1)
		    {
		      swap_bytes(buffer,1,4);
		    }
		  ptfloat = (float*)buffer;
		  timeVariableArray[i] = ptfloat[0];
		  fprintf(ascOutput,"%e,\t%f\n",timeVariableArray[i],dataPointArray[i]);
		}//end for - Second Segment of SAC received - Time stamps;  File Writing done as well
	    }

	}//end of else - that is, sacInput was not NULL

      exitFlag = 1;
   }/*End of Main While*/

  return;
}/*End of Main Program*/


/*Required Functions*/

/*Endian Checker*/
int endian_check() 
{
    int i = 1;
    char *p = (char *)&i;

    if (p[0] == 1)
        return LITTLE_ENDIAN;
    else
        return BIG_ENDIAN;
}

/*Byte Swapper*/
void swap_bytes(char *buffer,int samples,int size)
{
  char hold;
  int  i;

  for(i=0;i<samples*size;i=i+size)
  {
    if( size == 4 )
    {
      hold = buffer[i];
      buffer[i] = buffer[i+3];
      buffer[i+3] = hold;
      hold = buffer[i+1];
      buffer[i+1] = buffer[i+2];
      buffer[i+2] = hold;
    }
    else if( size == 2 )
    {
      hold = buffer[i];
      buffer[i] = buffer[i+1];
      buffer[i+1] = hold;
    }
  }
}
